package peer

import (
	"fmt"

	"cloud.o-forge.io/core/oc-lib/models/utils"
	"cloud.o-forge.io/core/oc-lib/tools"
)

// now write a go enum for the state partner with self, blacklist, partner

type PeerState int

const (
	NONE PeerState = iota
	SELF
	PARTNER
	BLACKLIST
)

func (m PeerState) String() string {
	return [...]string{"NONE", "SELF", "PARTNER", "BLACKLIST"}[m]
}

func (m PeerState) EnumIndex() int {
	return int(m)
}

// Peer is a struct that represents a peer
type Peer struct {
	utils.AbstractObject
	Url             string          `json:"url" bson:"url" validate:"required"`                       // Url is the URL of the peer (base64url)
	WalletAddress   string          `json:"wallet_address" bson:"wallet_address" validate:"required"` // WalletAddress is the wallet address of the peer
	PublicKey       string          `json:"public_key" bson:"public_key" validate:"required"`         // PublicKey is the public key of the peer
	State           PeerState       `json:"state" bson:"state" default:"0"`
	ServicesState   map[string]int  `json:"services_state,omitempty" bson:"services_state,omitempty"`
	FailedExecution []PeerExecution `json:"failed_execution" bson:"failed_execution"` // FailedExecution is the list of failed executions, to be retried
}

func (ao *Peer) VerifyAuth(request *tools.APIRequest) bool {
	return true
}

// AddExecution adds an execution to the list of failed executions
func (ao *Peer) AddExecution(exec PeerExecution) {
	found := false
	for _, v := range ao.FailedExecution { // Check if the execution is already in the list
		if v.Url == exec.Url && v.Method == exec.Method && fmt.Sprint(v.Body) == fmt.Sprint(exec.Body) {
			found = true
			break
		}
	}
	if !found {
		ao.FailedExecution = append(ao.FailedExecution, exec)
	}
}

// RemoveExecution removes an execution from the list of failed executions
func (ao *Peer) RemoveExecution(exec PeerExecution) {
	new := []PeerExecution{}
	for i, v := range ao.FailedExecution {
		if !(v.Url == exec.Url && v.Method == exec.Method && fmt.Sprint(v.Body) == fmt.Sprint(exec.Body)) {
			new = append(new, ao.FailedExecution[i])
		}
	}
	ao.FailedExecution = new
}

// IsMySelf checks if the peer is the local peer
func (p *Peer) IsMySelf() (bool, string) {
	d, code, err := NewAccessor(nil).Search(nil, SELF.String(), p.IsDraft)
	if code != 200 || err != nil || len(d) == 0 {
		return false, ""
	}
	id := d[0].GetID()
	return p.UUID == id, id
}

// LaunchPeerExecution launches an execution on a peer
func (p *Peer) LaunchPeerExecution(peerID string, dataID string, dt tools.DataType, method tools.METHOD, body interface{}, caller *tools.HTTPCaller) (*PeerExecution, error) {
	p.UUID = peerID
	return cache.LaunchPeerExecution(peerID, dataID, dt, method, body, caller) // Launch the execution on the peer through the cache
}
func (d *Peer) GetAccessor(request *tools.APIRequest) utils.Accessor {
	data := NewAccessor(request) // Create a new instance of the accessor
	return data
}

func (r *Peer) CanDelete() bool {
	return false // only draft order can be deleted
}
