package workflow_execution

import (
	"strings"
	"time"

	"cloud.o-forge.io/core/oc-lib/dbs"
	"cloud.o-forge.io/core/oc-lib/models/booking"
	"cloud.o-forge.io/core/oc-lib/models/common/enum"
	"cloud.o-forge.io/core/oc-lib/models/common/pricing"
	"cloud.o-forge.io/core/oc-lib/models/utils"
	"cloud.o-forge.io/core/oc-lib/tools"
	"github.com/google/uuid"
	"go.mongodb.org/mongo-driver/bson/primitive"
)

/*
* WorkflowExecution is a struct that represents a list of workflow executions
* Warning: No user can write (del, post, put) a workflow execution, it is only used by the system
* workflows generate their own executions
 */
type WorkflowExecution struct {
	utils.AbstractObject                                // AbstractObject contains the basic fields of an object (id, name)
	PeerBookByGraph      map[string]map[string][]string `json:"peer_book_by_graph,omitempty" bson:"peer_book_by_graph,omitempty"` // BookByResource is a map of the resource id and the list of the booking id
	ExecutionsID         string                         `json:"executions_id,omitempty" bson:"executions_id,omitempty"`
	ExecDate             time.Time                      `json:"execution_date,omitempty" bson:"execution_date,omitempty" validate:"required"` // ExecDate is the execution date of the workflow, is required
	EndDate              *time.Time                     `json:"end_date,omitempty" bson:"end_date,omitempty"`                                 // EndDate is the end date of the workflow
	State                enum.BookingStatus             `json:"state" bson:"state" default:"0"`                                               // TEMPORARY TODO DEFAULT 1 -> 0 State is the state of the workflow
	WorkflowID           string                         `json:"workflow_id" bson:"workflow_id,omitempty"`                                     // WorkflowID is the ID of the workflow
}

func (r *WorkflowExecution) StoreDraftDefault() {
	r.IsDraft = false // TODO: TEMPORARY
	r.State = enum.SCHEDULED
}

func (r *WorkflowExecution) CanUpdate(set utils.DBObject) (bool, utils.DBObject) {
	if r.State != set.(*WorkflowExecution).State {
		return true, &WorkflowExecution{State: set.(*WorkflowExecution).State} // only state can be updated
	}
	return !r.IsDraft, set // only draft buying can be updated
}

func (r *WorkflowExecution) CanDelete() bool {
	return r.IsDraft // only draft bookings can be deleted
}

func (wfa *WorkflowExecution) Equals(we *WorkflowExecution) bool {
	return wfa.ExecDate.Equal(we.ExecDate) && wfa.WorkflowID == we.WorkflowID
}

func (ws *WorkflowExecution) PurgeDraft(request *tools.APIRequest) error {
	if ws.EndDate == nil {
		// if no end... then Book like a savage
		e := ws.ExecDate.Add(time.Hour)
		ws.EndDate = &e
	}
	accessor := ws.GetAccessor(request)
	res, code, err := accessor.Search(&dbs.Filters{
		And: map[string][]dbs.Filter{ // check if there is a booking on the same compute resource by filtering on the compute_resource_id, the state and the execution date
			"state":       {{Operator: dbs.EQUAL.String(), Value: enum.DRAFT.EnumIndex()}},
			"workflow_id": {{Operator: dbs.EQUAL.String(), Value: ws.WorkflowID}},
			"execution_date": {
				{Operator: dbs.LTE.String(), Value: primitive.NewDateTimeFromTime(*ws.EndDate)},
				{Operator: dbs.GTE.String(), Value: primitive.NewDateTimeFromTime(ws.ExecDate)},
			},
		},
	}, "", ws.IsDraft)
	if code != 200 || err != nil {
		return err
	}
	for _, r := range res {
		accessor.DeleteOne(r.GetID())
	}
	return nil
}

// tool to transform the argo status to a state
func (wfa *WorkflowExecution) ArgoStatusToState(status string) *WorkflowExecution {
	status = strings.ToLower(status)
	switch status {
	case "succeeded": // Succeeded
		wfa.State = enum.SUCCESS
	case "pending": // Pending
		wfa.State = enum.SCHEDULED
	case "running": // Running
		wfa.State = enum.STARTED
	default: // Failed
		wfa.State = enum.FAILURE
	}
	return wfa
}

func (r *WorkflowExecution) GenerateID() {
	if r.UUID == "" {
		r.UUID = uuid.New().String()
	}
}

func (d *WorkflowExecution) GetName() string {
	return d.UUID + "_" + d.ExecDate.String()
}

func (d *WorkflowExecution) GetAccessor(request *tools.APIRequest) utils.Accessor {
	return NewAccessor(request) // Create a new instance of the accessor
}

func (d *WorkflowExecution) VerifyAuth(request *tools.APIRequest) bool {
	return true
}

func (d *WorkflowExecution) Book(executionsID string, wfID string, priceds map[tools.DataType]map[string]pricing.PricedItemITF) []*booking.Booking {
	booking := d.bookEach(executionsID, wfID, tools.STORAGE_RESOURCE, priceds[tools.STORAGE_RESOURCE])
	booking = append(booking, d.bookEach(executionsID, wfID, tools.PROCESSING_RESOURCE, priceds[tools.PROCESSING_RESOURCE])...)
	booking = append(booking,d.bookEach(executionsID, wfID, tools.COMPUTE_RESOURCE, priceds[tools.COMPUTE_RESOURCE])...)
	booking = append(booking,d.bookEach(executionsID, wfID, tools.DATA_RESOURCE, priceds[tools.DATA_RESOURCE])...)
	return booking
}

func (d *WorkflowExecution) bookEach(executionsID string, wfID string, dt tools.DataType, priceds map[string]pricing.PricedItemITF) []*booking.Booking {
	items := []*booking.Booking{}
	for itemID, priced := range priceds {
		if d.PeerBookByGraph == nil {
			d.PeerBookByGraph = map[string]map[string][]string{}
		}
		if d.PeerBookByGraph[priced.GetCreatorID()] == nil {
			d.PeerBookByGraph[priced.GetCreatorID()] = map[string][]string{}
		}
		if d.PeerBookByGraph[priced.GetCreatorID()][itemID] == nil {
			d.PeerBookByGraph[priced.GetCreatorID()][itemID] = []string{}
		}
		start := d.ExecDate
		if s := priced.GetLocationStart(); s != nil {
			start = *s
		}
		end := start.Add(time.Duration(priced.GetExplicitDurationInS()) * time.Second)
		bookingItem := &booking.Booking{
			AbstractObject: utils.AbstractObject{
				UUID: uuid.New().String(),
				Name: d.GetName() + "_" + executionsID + "_" + wfID,
			},
			ExecutionsID:      executionsID,
			State:             enum.SCHEDULED,
			ResourceID:        priced.GetID(),
			ResourceType:      dt,
			DestPeerID:        priced.GetCreatorID(),
			WorkflowID:        wfID,
			ExecutionID:       d.GetID(),
			ExpectedStartDate: start,
			ExpectedEndDate:   &end,
		}
		items = append(items, bookingItem)
		d.PeerBookByGraph[priced.GetCreatorID()][itemID] = append(
			d.PeerBookByGraph[priced.GetCreatorID()][itemID], bookingItem.GetID())
	}
	return items
}
